use std::{collections::HashSet, time::Duration};

use futures::StreamExt;
use log::warn;
use serenity::{
    collector::EventCollectorBuilder,
    model::prelude::{
        interaction::application_command::ApplicationCommandInteraction, Event, EventType, Message,
        ReactionType,
    },
    prelude::Context,
};

use crate::discord::commands::commons::CommandExecutionError;

use super::embed_builder::EmbedMessageBuilder;

const MAX_SELECTABLE_PER_PAGE: usize = 10;
const COLLECTOR_MAX_DURATION_SEC: u64 = 5000;

/*
 * Match the Discord 0 to  9 icon (that are encoded with three utf-8 character)
 * We can notice here that only the first character change between the emote.
 * This character is in fact the encoded number related to the emote.
 * Ex: '\u{0030}' = '0', \u{0031} = '1' ...
 */
const SELECTION_EMOTES: [&str; 10] = [
    "\u{0030}\u{FE0F}\u{20E3}",
    "\u{0031}\u{FE0F}\u{20E3}",
    "\u{0032}\u{FE0F}\u{20E3}",
    "\u{0033}\u{FE0F}\u{20E3}",
    "\u{0034}\u{FE0F}\u{20E3}",
    "\u{0035}\u{FE0F}\u{20E3}",
    "\u{0036}\u{FE0F}\u{20E3}",
    "\u{0037}\u{FE0F}\u{20E3}",
    "\u{0038}\u{FE0F}\u{20E3}",
    "\u{0039}\u{FE0F}\u{20E3}",
];

const PREVIOUS_PAGE_EMOTE: &str = "\u{2B05}";
const NEXT_PAGE_EMOTE: &str = "\u{27A1}";
const CONFIRM_EMOTE: &str = "\u{2705}";
const CANCEL_EMOTE: &str = "\u{274C}";

pub struct EmbedSelector<'a> {
    interaction: &'a ApplicationCommandInteraction,
    context: &'a Context,
    embed_answer: Option<Message>,
    title: String,
    description: String,
    selection: HashSet<String>,
    selectable: Vec<String>,
    page_number: usize,
    current_page: usize,
    aborted: bool,
}

impl<'a> EmbedSelector<'a> {
    pub fn new(
        title: String,
        description: String,
        interaction: &'a ApplicationCommandInteraction,
        context: &'a Context,
        selectable: Vec<String>,
        initial_selection: Option<HashSet<String>>,
    ) -> Self {
        let selection = match initial_selection {
            Some(r) => r,
            None => HashSet::new(),
        };

        let mut selector = EmbedSelector {
            interaction,
            context,
            embed_answer: None,
            title,
            description,
            selection,
            selectable: selectable.clone(),
            page_number: (selectable.len() as f32 / MAX_SELECTABLE_PER_PAGE as f32).ceil() as usize,
            current_page: 1,
            aborted: false,
        };

        selector.selectable.sort();
        selector
    }

    pub async fn get_user_selection(
        &mut self,
    ) -> Result<Option<HashSet<String>>, CommandExecutionError> {
        let embed_builder = EmbedMessageBuilder::new(self.context).await?;

        match self
            .interaction
            .edit_original_interaction_response(self.context, |response| {
                response.set_embed(embed_builder.create_selection_embed(
                    &self.title,
                    &self.description,
                    &self.selectable[0..self.get_current_page_choice_number()],
                    &self.selection,
                    self.page_number,
                    1,
                ))
            })
            .await
        {
            Ok(m) => {
                self.embed_answer = Some(m);
                self.refresh_reactions().await?;
                Ok(self.wait_selector_end().await?)
            }
            Err(_e) => Err(CommandExecutionError::DiscordAPICallError(
                "Failed to edit original interaction responnse".to_string(),
            )),
        }
    }

    async fn refresh_reactions(&self) -> Result<(), CommandExecutionError> {
        if let Some(answer) = &self.embed_answer {
            answer.delete_reactions(self.context).await.map_err(|_e| {
                CommandExecutionError::DiscordAPICallError(
                    "Failed to delete reaction on the current selector".to_string(),
                )
            })?;

            for emote in SELECTION_EMOTES[0..self.get_current_page_choice_number()]
                .iter()
                .chain(
                    [
                        PREVIOUS_PAGE_EMOTE,
                        NEXT_PAGE_EMOTE,
                        CANCEL_EMOTE,
                        CONFIRM_EMOTE,
                    ]
                    .iter(),
                )
            {
                match &self.embed_answer {
                    Some(a) => a
                        .react(self.context, ReactionType::Unicode(emote.to_string()))
                        .await
                        .map_err(|_e| {
                            CommandExecutionError::DiscordAPICallError(
                                "Failed to add reactions on the current selector".to_string(),
                            )
                        }),
                    None => Err(CommandExecutionError::SelectorError(
                        "Failed to refresh the reactions of the current selector".to_string(),
                    )),
                }?;
            }
            Ok(())
        } else {
            Err(CommandExecutionError::SelectorError(
                "Tried to delete reaction from a non existent message".to_string(),
            ))
        }
    }

    async fn wait_selector_end(
        &mut self,
    ) -> Result<Option<HashSet<String>>, CommandExecutionError> {
        let answer = match &self.embed_answer {
            Some(a) => Ok(a),
            None => Err(CommandExecutionError::SelectorError(
                "Tried to start collector before sending it".to_string(),
            )),
        }?;

        let mut collector = EventCollectorBuilder::new(self.context)
            .add_event_type(EventType::ReactionAdd)
            .add_user_id(self.interaction.user.id)
            .add_message_id(*answer.id.as_u64())
            .timeout(Duration::from_secs(COLLECTOR_MAX_DURATION_SEC))
            .build()
            .map_err(|_e| {
                CommandExecutionError::SelectorError(
                    "Failed to build the EventCollector".to_string(),
                )
            })?;

        while let Some(reaction_event) = collector.next().await {
            let reaction = match *reaction_event {
                Event::ReactionAdd(ref r) => &r.reaction,
                ref e => {
                    warn!("Received unexpected event in the selector EventCollector: {e:#?}");
                    continue;
                }
            };

            if let ReactionType::Unicode(reaction_str) = &reaction.emoji {
                match reaction_str.as_str() {
                    PREVIOUS_PAGE_EMOTE => self.previous_page().await?,
                    NEXT_PAGE_EMOTE => self.next_page().await?,
                    CANCEL_EMOTE => {
                        self.aborted = true;
                        break;
                    }
                    CONFIRM_EMOTE => {
                        break;
                    }
                    r => {
                        if SELECTION_EMOTES.contains(&reaction_str.as_str()) {
                            // Extract the number part of the emote unicode
                            let selected_nb = match r.chars().next() {
                                Some(c) => match c.to_digit(10) {
                                    Some(nb) => nb as usize,
                                    None => {
                                        warn!("Failed to cast emote code into number");
                                        continue;
                                    }
                                },
                                None => {
                                    warn!("Received empty emote code in ReactionAdd event");
                                    continue;
                                }
                            };

                            let tag_index: usize =
                                (self.current_page - 1) * MAX_SELECTABLE_PER_PAGE + selected_nb;

                            let tag = &self.selectable[tag_index];

                            if self.selection.contains(tag) {
                                self.selection.remove(tag);
                            } else {
                                self.selection.insert(tag.clone());
                            }
                        }

                        reaction.delete(self.context).await.map_err(|_e| {
                            CommandExecutionError::DiscordAPICallError(
                                "Failed to delete reaction from selector".to_string(),
                            )
                        })?;

                        self.refresh_embed_selection().await?;
                    }
                }
            }
        }
        collector.stop();

        if self.aborted {
            Ok(None)
        } else {
            Ok(Some(self.selection.clone()))
        }
    }

    async fn next_page(&mut self) -> Result<(), CommandExecutionError> {
        if self.current_page != self.page_number {
            self.current_page += 1;
            self.refresh_embed_selection().await?;
            self.refresh_reactions().await?;
        }
        Ok(())
    }

    async fn previous_page(&mut self) -> Result<(), CommandExecutionError> {
        if self.current_page != 1 {
            self.current_page -= 1;
            self.refresh_embed_selection().await?;
            self.refresh_reactions().await?;
        }
        Ok(())
    }

    fn get_current_page_choice_number(&self) -> usize {
        if self.page_number == self.current_page
            && self.selectable.len() % MAX_SELECTABLE_PER_PAGE != 0
        {
            self.selectable.len() % MAX_SELECTABLE_PER_PAGE
        } else {
            MAX_SELECTABLE_PER_PAGE
        }
    }

    async fn refresh_embed_selection(&mut self) -> Result<(), CommandExecutionError> {
        let embed_builder = EmbedMessageBuilder::new(self.context).await?;

        let curr_choices = self.selectable
            [(self.current_page - 1) * 10..(self.current_page * 10).min(self.selectable.len())]
            .to_vec();

        self.embed_answer
            .as_mut()
            .unwrap()
            .edit(self.context, |msg| {
                msg.set_embed(embed_builder.create_selection_embed(
                    &self.title,
                    &self.description,
                    &curr_choices,
                    &self.selection,
                    self.page_number,
                    self.current_page,
                ))
            })
            .await
            .map_err(|_e| {
                CommandExecutionError::DiscordAPICallError(
                    "Failed to edit selector content".to_string(),
                )
            })
    }
}
